from zope import interface

class IAuditable( interface.Interface ):
    """
    marker interface to apply auditing/object log feature
    """

class IRecorder( interface.Interface ):
    """
    a recorder for changes to an object
    """

    def objectAdded( object, event ):
        " "
    def objectModified( object, event ):
        " "
    def objectStateChanged( object, event ):
        " "
    def objectDeleted( object, event ):
        " "
    def objectNewVersion( object, event ):
        " "
    def objectRevertedVersion( object, event ):
        " "

    



"""
Auditing of Changes for Domain Objects.

implementation note, we dance around with attaching change ids to the
events, to do a crude coalesce in band to the event subscribers, as
workflow state changes are also modification events.
"""

import interfaces

def objectAdded( ob, event):
    auditor = interfaces.IRecorder( ob )
    auditor.objectAdded( event )
    
def objectModified( ob, event ):
    auditor = interfaces.IRecorder( ob )  
    if getattr( event, 'change_id', None):
        return
    auditor.objectModified(event )    
    
def objectDeleted( ob, event ):
    auditor = interfaces.IRecorder( ob )    
    auditor.objectDeleted( event )        

def objectStateChange( ob, event ):
    auditor = interfaces.IRecorder( ob )
    change_id = auditor.objectStateChanged( event )
    event.change_id = change_id
    
def objectNewVersion( ob, event ):
    auditor = interfaces.IRecorder( ob )
    if not getattr( event, 'change_id', None):
        change_id = auditor.objectNewVersion( event )
    else:
        change_id = event.change_id
    event.version.change_id = change_id

def objectRevertedVersion( ob, event ):
    # slightly obnoxious hand off between event handlers (objectnewV, objectrevertedV),
    # stuffing onto the event for value passing
    auditor = interfaces.IRecorder( ob )
    change_id = auditor.objectRevertedVersion( event )   
    event.change_id = change_id
    





from datetime import datetime
import sqlalchemy as rdb


def make_changes_table( table, metadata ):
    """ create an object log table for an object, this assumes per """
    table_name = table.name
    entity_name =  table_name.endswith('s') and table_name[:-1] or table_name
    
    changes_name = "%s_changes"%( entity_name )
    #import pdb; pdb.set_trace()

    fk_id = ("%s.%s"%(table.name, list(table.primary_key.columns)[0].name ))
    
    changes_table = rdb.Table(
            changes_name,
            metadata,
            rdb.Column( "change_id", rdb.Integer, primary_key=True ),
            rdb.Column( "content_id", rdb.Integer, rdb.ForeignKey( fk_id ) ),
            rdb.Column( "action", rdb.Unicode(16) ),
            rdb.Column( "date", rdb.Date, default=datetime.now),
            rdb.Column( "description", rdb.UnicodeText),
            rdb.Column( "notes", rdb.UnicodeText),
            rdb.Column( "user_id", rdb.Unicode(32) ) # Integer, rdb.ForeignKey('users.user_id') ),
    )
    
    return changes_table

def make_site_changes_table( metadata ):
    """ create a single changes table to capture for a whole site """



from zope import component
from zope.security.management import getInteraction
from zope.security.proxy import removeSecurityProxy
from zope.publisher.interfaces import IRequest

from zope import lifecycleevent
from datetime import datetime

from ore.alchemist.interfaces import IRelationChange
from ore.alchemist.model import queryModelInterface
from sqlalchemy import orm

import interfaces
from i18n import _

def provideRecorder( klass, change_table ):
    domain_interface = queryModelInterface( klass )
    recorder = type( "%sChangeRecorder"%(klass.__name__),
                     (ChangeRecorder,),
                     {'change_table':change_table} )
    component.provideAdapter( recorder, (domain_interface,), (interfaces.IRecorder,))
    
class ChangeRecorder( object ):

    change_table = None
    
    def __init__( self, context ):
        self.context = removeSecurityProxy(context)

    def objectAdded( self, event ):
        return self._objectChanged(u'added')
    
    def objectModified( self, event ):
        attrset =[]
        for attr in event.descriptions:
            if lifecycleevent.IAttributes.providedBy( attr ):
                attrset.extend(
                    [ attr.interface[a].title for a in attr.attributes]
                    )
            elif IRelationChange.providedBy(attr):
                attrset.append( attr.description )

        description = u", ".join( attrset )
        return self._objectChanged(u'modified', description )

    def objectStateChanged( self, event):
        description = _(u"""transition from %s to %s via %s - %s"""%( 
                        event.source,
                        event.destination,
                        event.transition.title,
                        event.comment ) )
        return self._objectChanged(u'workflow', description )
        
    def objectDeleted( self, event ):
        return self._objectChanged(u'deleted')

    def objectNewVersion( self, event ):
        return self._objectChanged(u"new-version", description=event.message )

    def objectRevertedVersion( self, event ):
        return self._objectChanged(u'reverted-version', description=event.message )
        
    def _objectChanged( self, change_kind, description=u'' ):
        oid, otype = self._getKey( self.context )
        user_id = self._getCurrentUserId()

        statement = self.change_table.insert(
            values = dict( action = change_kind,
                           date = datetime.now(),
                           user_id = user_id,
                           description = description,
                           content_type = otype,
                           content_id = oid )
            )
        value = statement.execute()
        return value.last_inserted_ids()[0]
        
    def _getKey( self, ob ):
        mapper = orm.object_mapper( ob )
        primary_key = mapper.primary_key_from_instance( ob )[0]
        return primary_key, unicode( ob.__class__.__name__ )

    def _getCurrentUserId( self ):
        interaction = getInteraction()
        for participation in interaction.participations:
            if IRequest.providedBy(participation):
                return participation.principal.id
        return None
    
    #raise RuntimeError(_("No IRequest in interaction"))    




from zope.publisher.browser import BrowserView
from zope.security.proxy import removeSecurityProxy
from zc.table import batching, column

import sqlalchemy as rdb
from sqlalchemy import orm

import interfaces

from i18n import _

class ChangeBaseView( BrowserView ):
    """
    base view for looking at changes to context
    """
    
    formatter_factory = batching.Formatter
    
    columns = [
        column.GetterColumn( title=_(u"action"), getter=lambda i,f:i['action'] ),    
        column.GetterColumn( title=_(u"date"), getter=lambda i,f: i['date'] ),
        column.GetterColumn( title=_(u"user"), getter=lambda i,f:i['user_id'] ),
        column.GetterColumn( title=_(u"description"), getter=lambda i,f:i['description'] ),
        ]    

    table_css = 'listing'
    
    def listing( self ):
        columns = self.columns
        formatter = self.formatter_factory( self.context,
                                            self.request,
                                            self.getFeedEntries(),
                                            prefix="results",
                                            visible_column_names = [c.name for c in columns],
                                            #sort_on = ('name', False)
                                            columns = columns )
        formatter.cssClasses['table'] = self.table_css
        formatter.updateBatching()
        return formatter()
        
    @property
    def _log_table( self ):
        auditor = interfaces.IRecorder( self.context )
        return auditor.change_table
        
    def getFeedEntries( self ):
        instance = removeSecurityProxy( self.context )        
        mapper = orm.object_mapper( instance )
        
        query = self._log_table.select().where(
            rdb.and_( self._log_table.c.content_id == rdb.bindparam('content_id') )
            )

        content_id = mapper.primary_key_from_instance( instance )[0] 
        content_changes = query.execute( content_id = content_id )
        return map( dict, content_changes)

class ChangeLog( ChangeBaseView ):
    """
    Change Log View for an object
    """
    form_name = "Change Log"



import os
from setuptools import setup, find_packages

def read(*rnames):
    return open(os.path.join(os.path.dirname(__file__), *rnames)).read()
    
setup(
    name="alchemist.audit",
    version="0.3.4",
    install_requires=['setuptools',
                      'ore.alchemist',
                      'zope.formlib',
                      'zc.table'],
    packages=find_packages(),
    namespace_packages=['alchemist'],
    package_data = {
    '': ['*.txt', '*.zcml', '*.pt'],
    },
    zip_safe=False,
    url="http://code.google.com/p/zope-alchemist",
    keywords="zope3 audit",
    author='Kapil Thangavelu',
    author_email='kapil.foss@gmail.com',
    description="Alchemist Auditing Components ( Event Subscribers, Change Recorders, UI ) for Relational Applications",

    long_description=(
        read('alchemist','audit','readme.txt')
        + '\n\n' +
        read('changes.txt')
        + '\n\n'
        ),
    classifiers=['Programming Language :: Python',
                 'Framework :: Zope3',
	         'Intended Audience :: Developers',
                 ],        
    license='ZPL',
    )

